@@ -11,6 +11,8 @@ import os
 import sys
 import decimal
 import datetime
+import uuid
+import json
 
 # third party
 import psycopg2
@@ -21,7 +23,14 @@ import psycopg2.extensions
 from ..base import SQLInterface, SQLConnection
 from ...compat import *
 from ...utils import get_objects
-from ...exception import UniqueError, InterfaceError
+from ...exception import (
+    InterfaceError,
+    UniqueError,
+    TableError,
+    FieldError,
+    UniqueError,
+    CloseError,
+)
 
 
 # class LoggingCursor(psycopg2.extras.RealDictCursor):
@@ -31,7 +40,54 @@ from ...exception import UniqueError, InterfaceError
 #         #psycopg2.extensions.cursor.execute(self, sql, args)
 
 
-#class Connection(psycopg2.extensions.connection, SQLConnection):
+class DictType(psycopg2.extensions.QuotedString):
+    """Converts from python dict to JSONB to be saved into the db
+
+    https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.QuotedString
+    https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.register_adapter
+
+    Surprisingly, Postgres converts the json value in the db back to dict just fine
+    but does not adapt a dict, I'm not sure why
+    """
+    @classmethod
+    def adapt(cls, val):
+        """adapter should be a function taking a single argument (the object to adapt)
+        and returning an object conforming to the ISQLQuote protocol (e.g. exposing
+        a getquoted() method). Once an object is registered, it can be safely used
+        in SQL queries and by the adapt() function
+
+        :param val: dict, the value coming from Python destined for Postgres
+        :returns: str
+        """
+        return cls(json.dumps(val))
+
+
+class StringType(object):
+    @classmethod
+    def convert(cls, val, cur):
+        """Convert a db string to a python str type
+
+        https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.new_type
+
+            These functions are used to manipulate type casters to convert from
+            PostgreSQL types to Python objects.
+
+            adapter should have signature fun(value, cur) where value is the string
+            representation returned by PostgreSQL and cur is the cursor from which
+            data are read. In case of NULL, value will be None. The adapter should
+            return the converted object.
+
+        :param val: str, the value coming from Postgres and destined for Python
+        :param cur: cursor
+        :returns: str
+        """
+        if isinstance(val, str) and val.startswith("\\x"):
+            buf = psycopg2.BINARY(val, cur)
+            val = bytes(buf).decode(cur.connection.encoding)
+
+        return val
+
+
 class Connection(SQLConnection, psycopg2.extensions.connection):
 #class Connection(SQLConnection, psycopg2.extras.LoggingConnection):
     """
@@ -44,45 +100,19 @@ class Connection(SQLConnection, psycopg2.extensions.connection):
         # http://initd.org/psycopg/docs/connection.html#connection.autocommit
         self.autocommit = True
 
-        if is_py2:
-            # unicode harden for python 2
-            # http://initd.org/psycopg/docs/usage.html#unicode-handling
-            psycopg2.extensions.register_type(psycopg2.extensions.UNICODE, self)
-            psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY, self)
+        psycopg2.extensions.register_type(
+            # https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.new_type
+            psycopg2.extensions.new_type(psycopg2.STRING.values, "STRING", StringType.convert)
+        )
 
-            # return ints for any long values, this normalizes with SQLite and 
-            # python 2.7+ transparently handles really large integer values
-            # with no long needed (long doesn't even exist in python 3+)
-            def normalize_long(v, cur):
-                return v if v is None else int(v)
-            psycopg2.extensions.register_type(
-                psycopg2.extensions.new_type(
-                    psycopg2.extensions.LONGINTEGER.values,
-                    b"LONGINTEGER",
-                    normalize_long
-                )
-            )
-
-        else:
-            def normalize_str(v, cur):
-                if isinstance(v, str) and v.startswith("\\x"):
-                    buf = psycopg2.BINARY(v, cur)
-                    v = bytes(buf).decode(cur.connection.encoding)
-
-                    #import binascii
-                    #pout.v(binascii.unhexlify(v[2:]))
-                    #v = v.encode(cur.connection.encoding)
-                    #v = bytes(v, encoding=cur.connection.encoding)
-                return v
-            psycopg2.extensions.register_type(
-                psycopg2.extensions.new_type(psycopg2.STRING.values, "STRING", normalize_str)
-            )
+        # https://www.psycopg.org/docs/extensions.html#psycopg2.extensions.register_adapter
+        psycopg2.extensions.register_adapter(dict, DictType.adapt)
 
 
         # http://initd.org/psycopg/docs/connection.html#connection.set_client_encoding
         # https://www.postgresql.org/docs/current/static/multibyte.html
         # > The default is the encoding defined by the database
-        # Not sure we want to override db encoding which is why I'm sure why I didn't
+        # Not sure we want to override db encoding which is probably why I didn't
         # set this previously
         #self.set_client_encoding("UTF8")
 
@@ -137,11 +167,13 @@ class PostgreSQL(SQLInterface):
 
     def free_connection(self, connection):
         if not self.connected: return
+
         if self._connection:
             self.log("freeing sync connection")
-            return
-        self.log("freeing async connection {}", id(connection))
-        self.connection_pool.putconn(connection)
+
+        else:
+            self.log("freeing async connection {}", id(connection))
+            self.connection_pool.putconn(connection)
 
     def get_connection(self):
         if not self.connected: self.connect()
@@ -185,24 +217,6 @@ class PostgreSQL(SQLInterface):
         # http://www.postgresql.org/message-id/CA+mi_8Y6UXtAmYKKBZAHBoY7F6giuT5WfE0wi3hR44XXYDsXzg@mail.gmail.com
         return [r['tablename'] for r in ret]
 
-    def _set_table(self, schema, **kwargs):
-        """
-        http://www.postgresql.org/docs/9.1/static/sql-createtable.html
-        http://www.postgresql.org/docs/8.1/static/datatype.html
-        http://pythonhosted.org/psycopg2/usage.html#adaptation-of-python-values-to-sql-types
-        """
-        query_str = []
-        query_str.append("CREATE TABLE {} (".format(self._normalize_table_name(schema)))
-
-        query_fields = []
-        for field_name, field in schema.fields.items():
-            query_fields.append('  {}'.format(self.get_field_SQL(field_name, field)))
-
-        query_str.append(",{}".format(os.linesep).join(query_fields))
-        query_str.append(')')
-        query_str = os.linesep.join(query_str)
-        ret = self.query(query_str, ignore_result=True, **kwargs)
-
     def _delete_table(self, schema, **kwargs):
         query_str = 'DROP TABLE IF EXISTS {} CASCADE'.format(self._normalize_table_name(schema))
         ret = self.query(query_str, ignore_result=True, **kwargs)
@@ -210,7 +224,7 @@ class PostgreSQL(SQLInterface):
     def _get_fields(self, table_name, **kwargs):
         """return all the fields for the given schema"""
         ret = {}
-        query_args = ['f', table_name]
+        query_args = ['f', self._normalize_table_name(table_name)]
 
         # I had to brush up on my join knowledge while writing this query
         # https://en.wikipedia.org/wiki/Join_(SQL)
@@ -256,6 +270,7 @@ class PostgreSQL(SQLInterface):
             "float4": float,
             "float8": float,
             "timestamp": datetime.datetime,
+            "timestamptz": datetime.datetime,
             "int2": int,
             "int4": int,
             "int8": long,
@@ -266,6 +281,7 @@ class PostgreSQL(SQLInterface):
             "bool": bool,
             "date": datetime.date,
             "blob": bytearray,
+            "jsonb": bytearray,
         }
 
         # the rows we can set: field_type, name, field_required, min_size, max_size,
@@ -317,60 +333,22 @@ class PostgreSQL(SQLInterface):
 
         return ret
 
-    def _set_index(self, schema, name, fields, **index_options):
+    def _set_index(self, schema, name, field_names, **index_options):
         """
         NOTE -- we set the index name using <table_name>_<name> format since indexes have to have
         a globally unique name in postgres
 
         http://www.postgresql.org/docs/9.1/static/sql-createindex.html
         """
-        index_fields = []
-        for field_name in fields:
-            field = schema.fields[field_name]
-            if issubclass(field.interface_type, basestring):
-                if field.options.get('ignore_case', False):
-                    field_name = 'UPPER({})'.format(self._normalize_name(field_name))
-            index_fields.append(field_name)
-
         query_str = 'CREATE {}INDEX {} ON {} USING BTREE ({})'.format(
             'UNIQUE ' if index_options.get('unique', False) else '',
             self._normalize_name("{}_{}".format(schema, name)),
             self._normalize_table_name(schema),
-            ', '.join(index_fields)
+            ', '.join(map(self._normalize_name, field_names))
         )
 
         return self.query(query_str, ignore_result=True, **index_options)
 
-    def _insert(self, schema, fields, **kwargs):
-        field_formats = []
-        field_names = []
-        query_vals = []
-        for field_name, field_val in fields.items():
-            field_names.append(self._normalize_name(field_name))
-            field_formats.append(self.val_placeholder)
-            query_vals.append(field_val)
-
-        pk_name = schema.pk_name
-        if pk_name:
-            query_str = 'INSERT INTO {} ({}) VALUES ({}) RETURNING {}'.format(
-                self._normalize_table_name(schema),
-                ', '.join(field_names),
-                ', '.join(field_formats),
-                self._normalize_name(pk_name),
-            )
-            ret = self.query(query_str, *query_vals, **kwargs)
-            ret = ret[0][pk_name]
-
-        else:
-            query_str = 'INSERT INTO {} ({}) VALUES ({})'.format(
-                self._normalize_table_name(schema),
-                ', '.join(field_names),
-                ', '.join(field_formats),
-            )
-            ret = self.query(query_str, *query_vals, ignore_result=True, **kwargs)
-
-        return ret
-
     def _normalize_field_SQL(self, schema, field_name, symbol):
         format_field_name = self._normalize_name(field_name)
         format_val_str = self.val_placeholder
@@ -378,11 +356,6 @@ class PostgreSQL(SQLInterface):
         if 'LIKE' in symbol:
             format_field_name += '::text'
 
-        # postgres specific for getting around case sensitivity:
-        if schema.fields[field_name].options.get('ignore_case', False):
-            format_field_name = 'UPPER({})'.format(field_name)
-            format_val_str = 'UPPER({})'.format(self.val_placeholder)
-
         return format_field_name, format_val_str
 
     def _normalize_sort_SQL(self, field_name, field_vals, sort_dir_str):
@@ -442,132 +415,109 @@ class PostgreSQL(SQLInterface):
             offset
         )
 
-    def get_field_SQL(self, field_name, field):
-        """
-        returns the SQL for a given field with full type information
-
-        field_name -- string -- the field's name
-        fields -- Field() -- the info for the field
-
-        return -- string -- the field type (eg, foo BOOL NOT NULL)
-        """
-        field_type = ""
-        interface_type = field.interface_type
-        is_pk = field.is_pk()
-
-        if issubclass(interface_type, bool):
-            field_type = 'BOOL'
-
-        elif issubclass(interface_type, long):
-            if is_pk:
-                field_type = 'BIGSERIAL PRIMARY KEY'
-            else:
-                field_type = 'BIGINT'
-
-        elif issubclass(interface_type, int):
-            #size = 2147483647
-            if is_pk:
-                field_type = 'BIGSERIAL PRIMARY KEY'
-
-            else:
-                size = sys.maxsize # http://stackoverflow.com/questions/7604966
-                if 'size' in field.options:
-                    size = field.options['size']
-                elif 'max_size' in field.options:
-                    size = field.options['max_size']
-
-                if size < 32767:
-                    field_type = 'SMALLINT'
-                else:
-                    if field.is_ref():
-                        field_type = 'BIGINT'
-                    else:
-                        field_type = 'INTEGER'
-
-        elif issubclass(interface_type, basestring):
-            fo = field.options
-            if field.is_ref():
-                ref_s = field.schema
-                fo = ref_s.pk.options
-
-            if 'size' in fo:
-                field_type = 'CHAR({})'.format(fo['size'])
-            elif 'max_size' in fo:
-                field_type = 'VARCHAR({})'.format(fo['max_size'])
-            else:
-                field_type = 'TEXT'
-
-            if is_pk:
-                field_type += ' PRIMARY KEY'
-
-        elif issubclass(interface_type, datetime.datetime):
-            # http://www.postgresql.org/docs/9.0/interactive/datatype-datetime.html
-            field_type = 'TIMESTAMP WITHOUT TIME ZONE'
-
-        elif issubclass(interface_type, datetime.date):
-            field_type = 'DATE'
-
-        elif issubclass(interface_type, float):
-            field_type = 'REAL'
-            size = field.options.get('size', field.options.get('max_size', 0))
-            if size > 6:
-                field_type = 'DOUBLE PRECISION'
-
-        elif issubclass(interface_type, decimal.Decimal):
-            field_type = 'NUMERIC'
-
-        elif issubclass(interface_type, bytearray):
-            field_type = 'BLOB'
+    def render_datatype_int_sql(self, field_name, field, **kwargs):
+        if field.is_pk():
+            field_type = 'BIGSERIAL PRIMARY KEY' # INT8
 
         else:
-            raise ValueError('unknown python type: {}'.format(interface_type.__name__))
-
-        if not is_pk:
-            if field.required:
-                field_type += ' NOT NULL'
-            else:
-                field_type += ' NULL'
-
             if field.is_ref():
-                ref_s = field.schema
-                if field.required: # strong ref, it deletes on fk row removal
-                    field_type += ' REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE CASCADE'.format(
-                        ref_s,
-                        ref_s.pk.name
-                    )
+                field_type = 'BIGINT' # INT8
 
-                else: # weak ref, it sets column to null on fk row removal
-                    field_type += ' REFERENCES {} ({}) ON UPDATE CASCADE ON DELETE SET NULL'.format(
-                        ref_s,
-                        ref_s.pk.name
-                    )
+            else:
+                # https://www.postgresql.org/docs/current/datatype-numeric.html
+                size_info = field.size_info()
+                size = size_info["size"]
 
-        return '{} {}'.format(self._normalize_name(field_name), field_type)
+                if size == 0:
+                    field_type = 'INTEGER' # INT4
 
-    def _handle_error(self, schema, e, **kwargs):
-        ret = False
+                elif size < 32767:
+                    field_type = 'SMALLINT' # INT2
 
+                elif size < 2147483647:
+                    field_type = 'INTEGER' # INT4
+
+                elif size < 9223372036854775807:
+                    field_type = 'BIGINT' # INT8
+
+                else:
+                    precision = size_info["precision"]
+                    field_type = f'NUMERIC({precision}, 0)'
+
+        return field_type
+
+    def render_datatype_str_sql(self, field_name, field, **kwargs):
+        if field.interface_options.get('ignore_case', False):
+            kwargs.setdefault("datatype", "CITEXT")
+
+        return super().render_datatype_str_sql(field_name, field, **kwargs)
+
+    def render_datatype_datetime_sql(self, field_name, field, **kwargs):
+        # http://www.postgresql.org/docs/9.0/interactive/datatype-datetime.html
+        #field_type = 'TIMESTAMP WITHOUT TIME ZONE'
+
+        # https://wiki.postgresql.org/wiki/Don't_Do_This#Don.27t_use_timestamp_.28without_time_zone.29
+        return 'TIMESTAMPTZ'
+
+    def render_datatype_dict_sql(self, field_name, field, **kwargs):
+        # https://www.postgresql.org/docs/current/datatype-json.html
+        # In general, most applications should prefer to store JSON data as
+        # jsonb, unless there are quite specialized needs
+        return 'JSONB'
+
+    def render_datatype_float_sql(self, field_name, field, **kwargs):
+        """
+        https://www.postgresql.org/docs/current/datatype-numeric.html#DATATYPE-NUMERIC-DECIMAL
+        """
+        size_info = field.size_info()
+        if size_info["has_precision"]:
+            precision = size_info["precision"]
+            scale = size_info["scale"]
+            field_type = f'NUMERIC({precision}, {scale})'
+
+        else:
+            size = size_info["size"]
+
+            # https://learn.microsoft.com/en-us/cpp/c-language/type-float
+            if size < 3.402823466e+38:
+                #field_type = 'REAL'
+                # REAL only has 6 digits after the period, double precision has
+                # 15 digits after the period
+                field_type = 'DOUBLE PRECISION'
+
+            elif size < 1.7976931348623158e+308:
+                field_type = 'DOUBLE PRECISION'
+
+            else:
+                precision = size_info["precision"]
+                field_type = f'NUMERIC({precision})'
+
+        return field_type
+
+    def render_datatype_uuid_sql(self, field_name, field, **kwargs):
+        # https://www.postgresql.org/docs/current/datatype-uuid.html
+        # https://www.postgresql.org/docs/current/functions-uuid.html
+        field_type = 'UUID'
+        if field.is_pk():
+            field_type += ' DEFAULT gen_random_uuid() PRIMARY KEY'
+        return field_type
+
+    def create_error(self, e, **kwargs):
         if isinstance(e, psycopg2.ProgrammingError):
             e_msg = String(e)
             if "does not exist" in e_msg:
                 if "column" in e_msg:
-                    #INSERT: 'column "cancelled_date" of relation "chat_followup" does not exist'
-                    #SELECT: 'column "cancelled_date" does not exist'
-                    try:
-                        ret = self._set_all_fields(schema, **kwargs)
-                    except ValueError as e:
-                        ret = False
+                    #INSERT: column "foo" of relation "<TABLE-NAME>" does not exist
+                    #SELECT: column "foo" does not exist
+                    e = FieldError(e)
 
                 else:
-                    #'relation "table_name" does not exit'
-                    ret = self._set_all_tables(schema, **kwargs)
+                    #'relation "<TABLE-NAME>" does not exit'
+                    e = TableError(e)
 
-        return ret
-
-    def _create_error(self, e, exc_info):
-        if isinstance(e, psycopg2.IntegrityError):
-            er = UniqueError(e, exc_info)
+        elif isinstance(e, psycopg2.IntegrityError):
+            e = UniqueError(e)
         else:
-            er = super(PostgreSQL, self)._create_error(e, exc_info)
-        return er
+            e = super().create_error(e, **kwargs)
+        return e
 
